#include "Bridge.h"

#include <QCoreApplication>
#include <QTcpSocket>

#include "HidInterface.h"

#include "def.h"

#include <QtDebug>

Bridge::Bridge(HidInterface* hid, QObject *parent)
    : QObject(parent)
    , m_hid(hid)
    , m_socket(0)
{
    connect(m_hid,      SIGNAL(readData(QByteArray)),
            this,         SLOT(hid_readData(QByteArray)));
    connect(m_hid,      SIGNAL(connectionLost()),
            this,         SLOT(hid_lostConnection()));
}

Bridge::~Bridge()
{
    if(m_socket != 0)
    {
        if(m_socket->isOpen())
            m_socket->close();
        delete m_socket;
    }
}

void Bridge::tcp_newConnection(QTcpSocket *socket)
{
    qDebug() << "Got new tcp connection";

    socket->setParent(this);
    socket->setSocketOption(QAbstractSocket::LowDelayOption, 1);

    if(m_socket != 0){
        qDebug() << "Error: already got a connection";
        return;
    }

    m_state = START_1;
    m_size = 0;

    m_socket = socket;
    connect(m_socket,   SIGNAL(readyRead()),
            this,         SLOT(tcp_readyRead()));
    connect(m_socket,   SIGNAL(disconnected()),
            this,         SLOT(tcp_lostConnection()));
    connect(m_socket,   SIGNAL(error(QAbstractSocket::SocketError)),
            this,         SLOT(tcp_error(QAbstractSocket::SocketError)));
}

void Bridge::tcp_lostConnection()
{
    qDebug() << "TCP - lost connection";
    m_socket->deleteLater();
    m_socket = 0;
}

void Bridge::tcp_readyRead()
{
    QByteArray data = m_socket->readAll();
    qDebug() << "    -> TCP - " << data.toHex();
    tcp_parsePackage(data);
}

void Bridge::tcp_error(QAbstractSocket::SocketError err)
{
    qDebug() << "Tcp Socket ERROR: " << err;
}

void Bridge::tcp_parsePackage( const QByteArray& package)
{
    QMutexLocker lock(&m_parseProtect);
    uchar byte = 0;
    for(int i = 0; i < package.size(); i++)
    {
        byte = package.at(i);
        switch(m_state)
        {
        case START_1:
            if(byte == 0xAA)
                m_state = START_2;
            break;
        case START_2:
            if(byte == 0x55)
                m_state = START_3;
            else
                m_state = START_1;
            break;
        case START_3:
            if(byte == 0xA5)
                m_state = LEN_1;
            else
                m_state = START_1;
            break;
        case LEN_1:
            m_size = 0;
            m_size += ((uint)byte) << 24;
            m_state = LEN_2;
            break;
        case LEN_2:
            m_size += ((uint)byte) << 16;
            m_state = LEN_3;
            break;
        case LEN_3:
            m_size += ((uint)byte) << 8;
            m_state = LEN_4;
            break;
        case LEN_4:
            m_size += ((uint)byte);
            m_state = DATA;
            m_data = QByteArray(m_size, (char)0);
            m_dataIndex = 0;
            break;
        case DATA:
            m_data[m_dataIndex++] = byte;
            if(m_dataIndex == m_size){
                qDebug() << "TCP -> HID - " << m_data.toHex();
                m_hid->write(m_data);
                m_state = START_1;
            }
            break;
        }
    }
}


void Bridge::hid_lostConnection()
{
    qDebug() << "HID - lost connection";
    m_hid->close();
    qApp->exit(RET_HID_LOST_CONNECTION);
}

void Bridge::hid_readData( QByteArray data )
{
    if(m_socket != 0){
        m_socket->write(data);
        m_socket->flush();
        qDebug() << "HID -> TCP - " << data.toHex();
    }
    else
        qDebug() << "Reade data, but could not forward..";
}
