#include "HidInterface.h"

#include <QtDebug>
#include <QTime>

#include <stdbool.h>
#include <hid.h>

#define PATH1               0xff000001
#define PATH2               0xff000001
#define PATHLEN             2

#define MAX_INPUT_DATA_LENGTH     1024
#define MAX_OUTPUT_DATA_LENGTH    300


HidInterface::HidInterface(quint16 vendorId, quint16 productId, uint endpoint, QObject *parent)
    : QThread(parent)
    , m_vendorId(vendorId)
    , m_productId(productId)
    , m_endpoint(endpoint)
    , m_active(false)
    , m_open(false)
    , m_hidInterface(0)
{
}

HidInterface::~HidInterface()
{
    if(m_active){
        m_active = false;
        while(!this->isFinished())
            msleep(100);
    }

    close();
}

bool HidInterface::open()
{
    HIDInterfaceMatcher matcher;
    matcher.vendor_id = m_vendorId;
    matcher.product_id = m_productId;
    matcher.matcher_fn = NULL;
    matcher.custom_data = NULL;
    matcher.custom_data_length = 0;

    hid_return ret;

    ret = hid_init();
    if(ret != HID_RET_SUCCESS)
    {
        return false;
    }

    /*
    // see include/debug.h for possible values
    hid_set_debug(HID_DEBUG_ALL);
    hid_set_debug_stream(stderr);
    // passed directly to libusb
    hid_set_usb_debug(HID_DEBUG_ALL);
    */


    m_hidInterface = hid_new_HIDInterface();
    if (m_hidInterface == 0)
    {
         hid_cleanup();
         return false;
    }

    ret = hid_force_open(m_hidInterface, 2, &matcher, 2);
    if (ret != HID_RET_SUCCESS)
    {
        hid_delete_HIDInterface(&m_hidInterface);
        hid_cleanup();
        return false;
    }

    m_open = true;
    this->start();

    return true;
}

void HidInterface::close ()
{
    if(m_hidInterface != 0)
    {
        if(m_open){
            hid_close(m_hidInterface);
            m_open = false;
        }

        hid_delete_HIDInterface(&m_hidInterface);
        m_hidInterface = 0;
    }

    if(hid_is_initialised())
        hid_cleanup();
}

bool HidInterface::write ( const QByteArray& data )
{
    static const int PATH_IN[PATHLEN] = { PATH1, PATH2 };

    qDebug() << "HID ->     - " << data.toHex();

    uchar   reportId;
    int     reportSize;
    int     reportCount = 0;
    bool    fitsIntoOneReport = true;

    do{

        fitsIntoOneReport = HidInterface::getOutputReportId(data.size(), reportId, reportSize);

        QByteArray buffer(reportSize + 1, (char)0x00);

        buffer[0] = reportId;
        buffer[1] = 0x00 | (fitsIntoOneReport)?0x00:0x02 | (reportCount++ == 0)?0x00:0x01;
        buffer.replace(2, data.size(), data);

        //qDebug() << "Trying to write " << buffer.size() << " bytes: " << buffer.toHex();

        qDebug() << "HID -> RAW - " << buffer.toHex();
        hid_return ret = hid_set_output_report(m_hidInterface, PATH_IN, PATHLEN, buffer, buffer.size());

        //qDebug() << "Write returned " << ret;

        if(ret != HID_RET_SUCCESS)
            return false;

    } while(!fitsIntoOneReport);

    return true;
}



void HidInterface::run ()
{
    uchar buffer[MAX_INPUT_DATA_LENGTH];
    QByteArray data;
    QTime time;
    int     msecsCount = 0;
    uchar   immediateReturnCount = 0;

    m_active = true;

    while(m_open && m_active)
    {
        memset(buffer, 0, MAX_INPUT_DATA_LENGTH);

        time.start();
        hid_return ret = hid_interrupt_read(m_hidInterface, m_endpoint, (char*)buffer, MAX_INPUT_DATA_LENGTH, 0);            // get current ReportTyp and PackageCount
        msecsCount = time.elapsed();

        if( (ret != HID_RET_SUCCESS && ret != 21 /*HID_RET_TIMEOUT*/)) // || time.elapsed() < 10 )
        {
            qWarning("Got unknown return code %d!!", ret);
            emit connectionLost();
            return;
        }

        //////////////////////////////////////////////////////////////
        // Report consits of:
        //  1 byte reportId that indicates the length
        //  N byte data that consist of
        //     1   byte link control byte
        //     N-1 byte real data

        uchar reportId = buffer[0];
        uchar linkCtrl = buffer[1];

        if(reportId == 0x00)
        {
            if(msecsCount == 0)
                immediateReturnCount++;

            qWarning("Received report Id 0x00, after %d msecs the %d. time", msecsCount, immediateReturnCount);

            if(immediateReturnCount > 100){
                qWarning("Returned %d times after 0 msecs!! -> Lost Connection!", immediateReturnCount);
                emit connectionLost();
                return;
            }

            continue;
        }


        const char* dataPtr = (const char*) &buffer[2];
        int dataSize = HidInterface::getInputReportSize(reportId) - 1;

        if(dataSize == 0)
            continue;

        immediateReturnCount = 0;

        QByteArray ba(dataPtr, dataSize); // remove reportId and linkCtrl

        if(linkCtrl == 0x00) // One and only package
        {
            qDebug() << "    -> HID - " << ba.toHex();
            emit readData(ba);
        }
        else if(linkCtrl == 0x02) // Start of a set of packages
        {
            data.clear();
            data.append(ba);
        }
        else if(linkCtrl == 0x03) // Mid package of a set
        {
            data.append(ba);
        }
        else if(linkCtrl == 0x01) // End package of a set
        {
            data.append(ba);
            qDebug() << "    -> HID - " << data.toHex();
            emit readData(data);
        }
    }
}



int HidInterface::getInputReportSize ( uchar reportId )
{
    switch(reportId)
    {
    case 0x01:
        return 5;
    case 0x02:
        return 9;
    case 0x03:
        return 13;
    case 0x04:
        return 17;
    case 0x05:
        return 25;
    case 0x06:
        return 49;
    case 0x07:
        return 95;
    case 0x08:
        return 193;
    case 0x09:
        return 257;
    case 0x0A:
        return 385;
    case 0x0B:
        return 513;
    case 0x0C:
        return 767;
    default:
        qWarning("Requested unknown report id");
        return 0;
    }
}

bool HidInterface::getOutputReportId ( int dataSize, uchar &reportId, int &reportSize )
{
    // Add one byte for the additional link control byte
    dataSize += 1;

    if(dataSize <= 5)
    {
        reportId = 0x0D;
        reportSize = 5;
    }
    else if(dataSize <= 9)
    {
        reportId = 0x0E;
        reportSize = 9;
    }
    else if(dataSize <= 13)
    {
        reportId = 0x0F;
        reportSize = 13;
    }
    else if(dataSize <= 17)
    {
        reportId = 0x10;
        reportSize = 17;
    }
    else if(dataSize <= 25)
    {
        reportId = 0x11;
        reportSize = 25;
    }
    else if(dataSize <= 49)
    {
        reportId = 0x12;
        reportSize = 49;
    }
    else if(dataSize <= 95)
    {
        reportId = 0x13;
        reportSize = 95;
    }
    else if(dataSize <= 193)
    {
        reportId = 0x14;
        reportSize = 193;
    }
    else if(dataSize <= 255)
    {
        reportId = 0x15;
        reportSize = 255;
    }
    else
    {
        reportId = 0x15;
        reportSize = 255;

        // Indicates that the data do NOT fit into one report
        return false;
    }

    // Indicates that the data fit into one report
    return true;
}
