#include <QtCore/QCoreApplication>
#include <QStringList>
#include <QSettings>
#include <QFile>
#include <QDir>
#include <QTextStream>
#include <QDateTime>

#include <QtDebug>

#include "TcpServer.h"
#include "HidInterface.h"
#include "Bridge.h"

#include "def.h"

#define DEBUG           0

#if DEBUG
#define DEBUG_TO_STDERR     0
#define DEBUG_TO_FILE       1
#else
#define DEBUG_TO_STDERR     0
#define DEBUG_TO_FILE       0
#endif

#if DEBUG_TO_FILE
QTextStream* trcFile = 0;
#endif

#if DEBUG
void myMessageOutput(QtMsgType type, const char *msg)
#else
void myMessageOutput(QtMsgType /*type*/, const char */*msg*/)
#endif
{
#if DEBUG
     switch (type) {
     case QtDebugMsg:
         fprintf(stderr, "%s\n", msg);
#if DEBUG_TO_FILE
         if(trcFile){
             *trcFile << QTime::currentTime().toString("hh:mm:ss.zzz") << " " << msg << "\n";
             trcFile->flush();
         }
#endif
         break;
     case QtWarningMsg:
         fprintf(stderr, "Warning: %s\n", msg);
#if DEBUG_TO_FILE
         if(trcFile){
             *trcFile << QTime::currentTime().toString("hh:mm:ss.zzz") << " Warning: " << msg << "\n";
             trcFile->flush();
         }
#endif
         break;
     case QtCriticalMsg:
         fprintf(stderr, "Critical: %s\n", msg);
#if DEBUG_TO_FILE
         if(trcFile){
             *trcFile << QTime::currentTime().toString("hh:mm:ss.zzz") << " Critical: " << msg << "\n";
             trcFile->flush();
         }
#endif
         break;
     case QtFatalMsg:
         fprintf(stderr, "Fatal: %s\n", msg);
#if DEBUG_TO_FILE
         if(trcFile){
             *trcFile << QTime::currentTime().toString("hh:mm:ss.zzz") << " Fatal: " << msg << "\n";
             trcFile->flush();
         }
#endif
         abort();
     }
#endif
}

void writeDefaultSettings(QString settingsFileName)
{
    QSettings settings(settingsFileName, QSettings::IniFormat);
    settings.beginGroup("HID-ReadEndpoints");
    settings.setValue("1209", "82");
    settings.setValue("120A", "82");
    settings.setValue("1290", "83");
    settings.endGroup();
}

int endpointNumberForProduct(QString settingsFileName, quint16 pid)
{
    int endpoint = 0x83;

    // Read settings of this product
    qDebug() << "Looking up settings @ " << settingsFileName;
    QSettings settings(settingsFileName, QSettings::IniFormat);
    settings.beginGroup("HID-ReadEndpoints");
    QStringList productIds = settings.childKeys();
    foreach(QString productId, productIds)
    {
        if(productId.compare(QString("%1").arg(pid, 4, 16, QChar('0')), Qt::CaseInsensitive) == 0)
        {
            QString ep = settings.value(productId, QVariant("83")).toString();
            bool ok = false;
            uint epShort = ep.toUInt(&ok, 16);
            endpoint = (uint) epShort;
            break;
        }
    }
    settings.endGroup();
    settings.sync();

    return endpoint;
}


int main(int argc, char *argv[])
{
    qInstallMsgHandler(myMessageOutput);
    QCoreApplication a(argc, argv);

#if DEBUG_TO_FILE
    QFile f(QCoreApplication::applicationDirPath() + "/trcFile.txt");
    if(f.open(QFile::WriteOnly | QFile::Truncate))
    {
        qDebug() << "File opened";
        trcFile = new QTextStream(&f);
    }
#endif

    QStringList arguments = a.arguments();

    // Variables that hold the vendor and product id of the device that should be opened
    quint16 vid = 0;
    quint16 pid = 0;

    foreach(QString str, arguments)
    {
        // Get vendor id, should look this: -vid05AC
        if(str.startsWith("-vid"))
        {
            str.remove(0, 4);               // remove "-vid"
            bool ok = false;
            vid = str.toInt(&ok, 16);       // try to cast vid
            if(!ok)
                vid = 0;
        }

        // Get product id, should look this: -pid1290
        if(str.startsWith("-pid"))
        {
            str.remove(0, 4);               // remove "-pid"
            bool ok = false;
            pid = str.toInt(&ok, 16);       // try to cast vid
            if(!ok)
                pid = 0;
        }
    }

    //qDebug() << "Got informations to start: ";
    //qDebug() << "  VID = " << vid;
    //qDebug() << "  PID = " << pid;

    // Check if all parameters were received
    if(vid == 0 || pid == 0)
    {
        qDebug() << "Did not get vendor or product id -> can not start";
        return RET_INVALID_PARAMS;
    }

    QString settingsFileName = QCoreApplication::applicationDirPath() + "/iPodHidSettings.ini";
    if(!QFile::exists(settingsFileName))
        writeDefaultSettings(settingsFileName);

    uint endpoint = endpointNumberForProduct(settingsFileName, pid);
    qDebug() << "Using endpoint " << endpoint << " for device";

    // Create HID interface with the given parameters
    HidInterface hid(vid, pid, endpoint);
    if(!hid.open())
    {
        qDebug() << "Could not open hid device";
        return RET_COULD_NOT_OPEN_HID;
    }

    // Create the bridge instance, that passes through data
    Bridge* bridge = new Bridge(&hid);

    // Create Tcp Server and start listening to any port
    TcpServer* server = new TcpServer();
    if(!server->listen())
    {
        qDebug() << "Could not start tcp server";
        return RET_COULD_NOT_START_TCP;
    }

    // Connect Tcp Server to bridge, so bridge can receive new connections
    QObject::connect(server,   SIGNAL(newTcpConnection(QTcpSocket*)),
                     bridge,     SLOT(tcp_newConnection(QTcpSocket*)));

    // Write the listening port to stdout out, so calling app can connect to this port
    qDebug() << "Output:";
    fprintf(stdout, "|%s|", QString("%1").arg(server->serverPort(), 4, 16, QChar('0')).toAscii().data());
    fflush(stdout);
    qDebug() << "Tcp Server running on port: " << server->serverPort();

    // Start event loop
    a.exec();

    delete bridge;
    delete server;

#if DEBUG_TO_FILE
    if(trcFile)
        delete trcFile;
    if(f.isOpen())
        f.close();
#endif

    return RET_HID_LOST_CONNECTION;
}
